using System;
using System.Data;
using System.Web.UI.WebControls;

using CMS.PortalControls;
using CMS.Helpers;

using CMS.DocumentEngine;
using CMS.Membership;
using CMS.SiteProvider;
using System.Collections.Generic;
using System.Net;
using System.Net.Http;
using System.Text;
using Custom;

namespace Custom
{
    public partial class AzureSearch : CMSAbstractWebPart
    {
        #region "Constants"

        private static Uri _serviceUri;
        private static HttpClient _httpClient;

        #endregion

        #region "Public properties"

        /// <summary>
        /// Gets or sets azure search service name
        /// </summary>
        public string AzureSearchServiceName
        {
            get
            {
                return ValidationHelper.GetString(GetValue("AzureSearchServiceName"), "");
            }
            set
            {
                SetValue("AzureSearchServiceName", value);
            }
        }

        /// <summary>
        /// Gets or sets azure search service key
        /// </summary>
        public string AzureSearchServiceKey
        {
            get
            {
                return ValidationHelper.GetString(GetValue("AzureSearchServiceKey"), "");
            }
            set
            {
                SetValue("AzureSearchServiceKey", value);
            }
        }

        /// <summary>
        /// Gets or sets azure search service key
        /// </summary>
        public string AzureSearchServiceIndexName
        {
            get
            {
                return ValidationHelper.GetString(GetValue("AzureSearchServiceIndexName"), "");
            }
            set
            {
                SetValue("AzureSearchServiceIndexName", value);
            }
        }



        #region "Methods"

        /// <summary>
        /// Content loaded event handler.
        /// </summary>
        public override void OnContentLoaded()
        {
            base.OnContentLoaded();
            SetupControl();
        }


        /// <summary>
        /// Initializes the control properties.
        /// </summary>
        protected void SetupControl()
        {
            if (this.StopProcessing)
            {
                // Do not process
            }
            else
            {
                //Build up the Azure Service path
                _serviceUri = new Uri("https://" + this.AzureSearchServiceName + ".search.windows.net");
                _httpClient = new HttpClient();
                _httpClient.DefaultRequestHeaders.Add("api-key", this.AzureSearchServiceKey);
            }
        }


        /// <summary>
        /// Reloads the control data.
        /// </summary>
        public override void ReloadData()
        {
            base.ReloadData();

            SetupControl();
        }

        protected void bt_Click(object sender, EventArgs e)
        {
            lblFacets.Text = "";
            lblResults.Text = "";
            lblTotal.Text = "";
            try
            {
                StringBuilder sb = new StringBuilder();
                Button btn = sender as Button;
                switch (btn.ID)
                {
                    case "btnDates":
                        // Create an instance of the Tree provider first
                        TreeProvider tree = new TreeProvider(MembershipContext.AuthenticatedUser);

                        // Fill dataset with documents
                        DataSet documents = tree.SelectNodes(SiteContext.CurrentSiteName, "/quotes/%", "en-us", false, "custom.quote");

                        if (!DataHelper.DataSourceIsEmpty(documents))
                        {
                            // Loop through all documents
                            foreach (DataRow documentRow in documents.Tables[0].Rows)
                            {
                                // Create a new Tree node from the data row
                                CMS.DocumentEngine.TreeNode editDocument = CMS.DocumentEngine.TreeNode.New("custom.quote", documentRow, tree);

                                DateTime dtStart = new DateTime(1990, 1, 1);
                                TimeSpan timeSpan = DateTime.Now - dtStart;
                                var randomTest = new Random();

                                TimeSpan newSpan = new TimeSpan(0, randomTest.Next(0, (int)timeSpan.TotalMinutes), 0);
                                DateTime newDate = dtStart + newSpan;

                                // Change coupled data
                                editDocument.SetValue("QuoteDate", newDate);

                                // Save to database
                                editDocument.Update();
                            }
                        }

                        break;
                    case "btnCreateIndex": //Create the index in Azure Saerch
                        Uri uri = new Uri(_serviceUri, "/indexes/" + this.AzureSearchServiceIndexName);
                        HttpResponseMessage response = AzureSearchHelper.SendSearchRequest(_httpClient, HttpMethod.Get, uri);
                        if (response.StatusCode == HttpStatusCode.NotFound)
                        {
                            sb.Append(CreateIndex());
                        }
                        else
                        {
                            sb.Append("<span class=\"errormessage\">Index already exists!</span>");
                        }
                        break;
                    case "btnLoadIndex": //Populate the Azure Search Index
                        sb.Append(LoadIndex());
                        break;
                    case "btnDeleteIndex": //Delete the index
                        sb.Append(DeleteIndex());
                        break;
                    case "btnSearch": //Search against the Azure Search
                        sb.Append(SearchIndex(ValidationHelper.GetString(tbSearch.Text, ""), cbScore.Checked, cbHighlight.Checked, cbFacets.Checked, ValidationHelper.GetString(hidFacet.Value, ""), ValidationHelper.GetString(ddlScoreRule.SelectedValue, "")));
                        hidFacet.Value = "";
                        lblSuggestions.Text = "";
                        break;
                    case "btnReset": //Do a whole lot of nothing, but make things look nice.
                        tbSearch.Text = "";
                        lblSuggestions.Text = "";
                        sb.Clear();
                        cbHighlight.Checked = false;
                        cbFacets.Checked = false;
                        cbScore.Checked = false;
                        ddlScoreRule.SelectedIndex = -1;
                        break;
                }
                lblResults.Text = sb.ToString();
            }
            catch (Exception ex)
            {
                lblResults.Text = "<span class=\"responsemessage\">" + ex.Message + "</span>";
            }
        }

        /// <summary>
        /// This function will create the specified index.
        /// </summary>
        /// <returns>string - Results response</returns>
        private string CreateIndex()
        {
            try
            {
                var definition = new
                {
                    Name = this.AzureSearchServiceIndexName,
                    Fields = new[]
                    {
                    new { Name = "DocumentID",Type = "Edm.String",Key = true,  Searchable = false, Filterable = false, Sortable = false, Facetable = false, Retrievable = true,  Suggestions = false },
                    new { Name = "NodeAliasPath",Type = "Edm.String",Key = false,  Searchable = false, Filterable = false, Sortable = false, Facetable = false, Retrievable = true,  Suggestions = false },
                    new { Name = "QuoteAuthor",Type = "Edm.String",Key = false,  Searchable = true, Filterable = true, Sortable = true, Facetable = true, Retrievable = true,  Suggestions = true },
                    new { Name = "QuoteText",Type = "Edm.String",Key = false, Searchable = true,  Filterable = true, Sortable = true,  Facetable = true, Retrievable = true,  Suggestions = true  },
                    new { Name = "QuoteDate",Type = "Edm.String",Key = false, Searchable = true,  Filterable = true, Sortable = true,  Facetable = true, Retrievable = true,  Suggestions = true  }
                }
                };

                Uri uri = new Uri(_serviceUri, "/indexes");
                string json = AzureSearchHelper.SerializeJson(definition);
                HttpResponseMessage response = AzureSearchHelper.SendSearchRequest(_httpClient, HttpMethod.Post, uri, json);
                response.EnsureSuccessStatusCode();

                return "<span class=\"responsemessage\">Index created.</span>";
            }
            catch (Exception ex)
            {
                return "<span class=\"errormessage\">Index not created.<br />" + ex.Message + "</span>";
            }
        }

        /// <summary>
        /// This function will delete the specified index.
        /// </summary>
        /// <returns>string - Results response</returns>
        private string DeleteIndex()
        {
            try
            {
                Uri uri = new Uri(_serviceUri, "/indexes/quotes");
                HttpResponseMessage response = AzureSearchHelper.SendSearchRequest(_httpClient, HttpMethod.Delete, uri);
                response.EnsureSuccessStatusCode();

                return "<span class=\"responsemessage\">Index deleted.</span>";
            }
            catch (Exception ex)
            {
                return "<span class=\"errormessage\">Index not deleted.<br />" + ex.Message + "</span>";
            }
        }

        /// <summary>
        /// This function will load the index with documents for the current site
        /// </summary>
        /// <returns>string - Results response.</returns>
        private string LoadIndex()
        {
            StringBuilder sb = new StringBuilder();
            try
            {
                int i = 1;

                // Create an instance of the Tree provider first
                TreeProvider tree = new TreeProvider(MembershipContext.AuthenticatedUser);

                sb.Append("{");
                sb.Append("\"value\": [");


                // Fill dataset with documents
                DataSet documents = tree.SelectNodes(SiteContext.CurrentSiteName, "/quotes/%", "en-us", false, "custom.quote");

                if (!DataHelper.DataSourceIsEmpty(documents))
                {
                    // Loop through all documents
                    foreach (DataRow dr in documents.Tables[0].Rows)
                    {
                        // Create a new Tree node from the data row
                        CMS.DocumentEngine.TreeNode doc = CMS.DocumentEngine.TreeNode.New("custom.quote", dr, tree);

                        sb.Append("{");
                        sb.Append("\"@search.action\":\"mergeOrUpload\",");
                        sb.Append("\"DocumentID\":\"" + doc.DocumentID + "\",");
                        sb.Append("\"NodeAliasPath\":\"" + doc.NodeAliasPath + "\",");
                        sb.Append("\"QuoteAuthor\":\"" + doc.GetValue("QuoteAuthor").ToString() + "\",");
                        sb.Append("\"QuoteText\":\"" + doc.GetValue("QuoteText").ToString().Replace("'", "''").Replace("\"", "''") + "\",");
                        sb.Append("\"QuoteDate\":\"" + doc.GetValue("QuoteDate").ToString() + "\"");
                        sb.Append("}");
                        if (i < documents.Tables[0].Rows.Count)
                        {
                            sb.Append(",");
                        }
                        i += 1;
                    }
                }


                Uri uri = new Uri(_serviceUri, "/indexes/" + this.AzureSearchServiceIndexName + "/docs/index");
                sb.Append("]");
                sb.Append("}");
                string json = sb.ToString();


                HttpResponseMessage response = AzureSearchHelper.SendSearchRequest(_httpClient, HttpMethod.Post, uri, json);
                response.EnsureSuccessStatusCode();

                return "<span class=\"responsemessage\">Index data loaded</span>";
            }
            catch (Exception ex)
            {
                return "<span class=\"errormessage\">Index data not created.<br />" + ex.Message + "</span>";
            }
        }

        /// <summary>
        /// This function will return a formatted string of the documents mathcing the specified search value
        /// </summary>
        /// <param name="strValue">string - Search value</param>
        /// <returns>string - Some totally awesome search results</returns>
        private string SearchIndex(string strValue, bool blnScore, bool blnHighlight, bool blnFacets, string strFacet, string strScoringRule)
        {
            StringBuilder sb = new StringBuilder();
            StringBuilder sbUri = new StringBuilder();
            try
            {
                //Build up the search parameter
                string search = "";
                search = "&search=" + Uri.EscapeDataString(strValue) + "*";

                //Get the Azure Search records for the specified value
                if (strValue.Length > 2)
                {
                    sbUri.Append("/indexes/" + this.AzureSearchServiceIndexName + "/docs?$count=true&orderby=QuoteText");
                    if (blnHighlight)
                    {
                        sbUri.Append("&highlight=QuoteAuthor,QuoteText&highlightPreTag=<span class='highlight'>&highlightPostTag=</span>");
                    }
                    if (blnFacets)
                    {
                        sbUri.Append("&facet=QuoteAuthor");
                    }
                    if (strFacet != "")
                    {
                        sbUri.Append("&$filter=QuoteAuthor eq '" + strFacet + "'");
                    }
                    if (strScoringRule != "")
                    {
                        sbUri.Append("&scoringProfile=" + ddlScoreRule.SelectedValue);
                    }
                    sbUri.Append(search);
                    Uri uri = new Uri(_serviceUri, sbUri.ToString());
                    HttpResponseMessage response = AzureSearchHelper.SendSearchRequest(_httpClient, HttpMethod.Get, uri);
                    AzureSearchHelper.EnsureSuccessfulSearchResponse(response);

                    dynamic results = AzureSearchHelper.DeserializeJson<dynamic>(response.Content.ReadAsStringAsync().Result);

                    lblTotal.Text = "Total Records: " + results["@odata.count"];

                    //Get the facets
                    if (results["@search.facets"] != null)
                    {
                        StringBuilder sbFacets = new StringBuilder();
                        foreach (var item in results["@search.facets"]["QuoteAuthor"])
                        {
                            sbFacets.Append("<a style=\"cursor: pointer;\" onclick=\"FacetClick('" + item["value"] + "');\">" + item["value"] + " (" + item["count"] + ")<br />");

                        }

                        //Add in an all link for the facets
                        sbFacets.Append("<a style=\"cursor: pointer;\" onclick=\"FacetClick('All');\">All<br />");
                        lblFacets.Text = sbFacets.ToString();
                    }


                    foreach (var item in results["value"])
                    {
                        if (blnScore)
                        {
                            if (item["@search.score"] != null)
                            {
                                sb.Append("Score: " + item["@search.score"].ToString() + "<br/>");
                            }
                        }
                        sb.Append("<a href=\"~" + item["NodeAliasPath"].ToString() + "\">");

                        if (item["@search.highlights"] != null)
                        {
                            if (item["@search.highlights"]["QuoteAuthor"] != null)
                            {
                                sb.Append(item["@search.highlights"]["QuoteAuthor"][0].Value + " - ");
                            }
                            else
                            {
                                sb.Append(item["QuoteAuthor"].ToString() + " - ");
                            }

                            if (item["@search.highlights"]["QuoteText"] != null)
                            {
                                sb.Append(item["@search.highlights"]["QuoteText"][0].Value);
                            }
                            else
                            {
                                sb.Append(item["QuoteText"].ToString().Replace("''", "'"));
                            }
                        }
                        else
                        {
                            sb.Append(item["QuoteAuthor"].ToString() + " - ");
                            sb.Append(item["QuoteText"].ToString().Replace("''", "'"));
                            sb.Append(" - " + item["QuoteDate"].ToString());
                        }
                        sb.Append("</a><br /><br />");
                    }
                }
                else
                {
                    sb.Append("<span class=\"errormessage\">You must enter atleast 3 characters.</span>");
                }
            }
            catch (Exception ex)
            {
                sb.Append("<span class=\"errormessage\">" + ex.Message + "</span>");
            }

            hidFacet.Value = "";

            return sb.ToString();

        }

        #endregion


        protected void tbSearch_TextChanged(object sender, EventArgs e)
        {
            TextBox tb = (TextBox)sender;
            StringBuilder sbSuggestions = new StringBuilder();
            StringBuilder sb = new StringBuilder();
            StringBuilder sbUri = new StringBuilder();
            try
            {
                if (tb.Text != "")
                {
                    if (tb.Text.Length > 2)
                    {
                        //Create a string for the suggestions
                        sbSuggestions.Append("<strong>Suggestions</strong>:  ");

                        //Build up the search parameter
                        sbUri.Append("/indexes/" + this.AzureSearchServiceIndexName + "/docs");
                        string search = "";
                        search = "/suggest/?search=" + Uri.EscapeDataString(tb.Text) + "&fuzzy=true&searchFields=QuoteAuthor";

                        sbUri.Append(search);
                        Uri uri = new Uri(_serviceUri, sbUri.ToString());
                        HttpResponseMessage response = AzureSearchHelper.SendSearchRequest(_httpClient, HttpMethod.Get, uri);
                        AzureSearchHelper.EnsureSuccessfulSearchResponse(response);

                        dynamic results = AzureSearchHelper.DeserializeJson<dynamic>(response.Content.ReadAsStringAsync().Result);


                        List<string> lstAuthors = new List<string>();

                        //Add the found suggested authors to the list
                        foreach (var item in results["value"])
                        {
                            if (!lstAuthors.Contains(item["@search.text"].ToString()))
                            {
                                lstAuthors.Add(item["@search.text"].ToString());
                            }
                        }

                        lstAuthors.Sort();

                        //Build up a link for each suggested author
                        foreach (string strAuthor in lstAuthors)
                        {
                            StringBuilder sbLink = new StringBuilder();
                            sbSuggestions.Append("<a href=\"javascript: selectauthor('");
                            sbSuggestions.Append(strAuthor);
                            sbSuggestions.Append("');\">");
                            sbSuggestions.Append(strAuthor + "&nbsp;&nbsp;");
                            sbSuggestions.Append("</a>  ");
                        }
                    }
                }
                lblSuggestions.Text = sbSuggestions.ToString();
            }
            catch (Exception ex)
            {
                sb.Append("<span class=\"errormessage\">" + ex.Message + "</span>");
            }
        }

        #endregion
    }
}